/* 
 * $Id: ripconf.c,v 1.2 1998/09/04 20:38:21 masaki Exp $
 */

#include <mrt.h>
#include <rip.h>
#include <config_file.h>
#include <protoconf.h>

/* Comments...
 * Config can work from a file or from user configuration via a
 * socket.
 */

int rip2_show (uii_connection_t *uii) { 
	return (rip_show (PROTO_RIP, uii)); }
int rip2_show_routing_table (uii_connection_t *uii) {
	return (rip_show_routing_table (PROTO_RIP, uii)); }
#ifdef HAVE_IPV6
int ripng_show (uii_connection_t *uii) { 
	return (rip_show (PROTO_RIPNG, uii)); }
int ripng_show_routing_table (uii_connection_t *uii) {
	return (rip_show_routing_table (PROTO_RIPNG, uii)); }
#endif /* HAVE_IPV6 */


typedef struct _dlist_t {
    int num;
    char *in_or_out;
    interface_t *interface;
} dlist_t;


static int
config_distribute_list (uii_connection_t * uii,
			int num, char *in_or_out, int opt_num, char *name)
{
    interface_t *interface;
    dlist_t *dlist;
    rip_t *rip = RIP;

#ifdef HAVE_IPV6
    if (CONFIG_MRTD->protocol == PROTO_RIPNG)
	rip = RIPNG;
#endif /* HAVE_IPV6 */

    if (num < 0 && num >= MAX_ALIST) {
	config_notice (TR_ERROR, uii,
		       "invalid access-list number (%d)\n", num);
	Delete (in_or_out);
	if (opt_num > 0)
	    Delete (name);
	return (-1);
    }

    if (opt_num <= 0)
	name = "";
    config_notice (TR_PARSE, uii,
		 "CONFIG distribute-list %d %s %s\n", num, in_or_out, name);

    if (*name) {
	if ((interface = find_interface_byname (name)) == NULL) {
	    config_notice (TR_ERROR, uii,
			   "unkown interface %s\n", name);
	    Delete (in_or_out);
	    Delete (name);
	    return (-1);
	}
	if (strcasecmp (in_or_out, "in") == 0)
	    interface->dlist_in[CONFIG_MRTD->protocol] = num;
	else
	    interface->dlist_out[CONFIG_MRTD->protocol] = num;

	dlist = New (dlist_t);
	dlist->num = num;
	dlist->in_or_out = in_or_out;
	dlist->interface = interface;
	LL_Add (rip->ll_dlists, dlist);
	Delete (name);
    }
    else {
	LL_Iterate (INTERFACE_MASTER->ll_interfaces, interface) {
	    if (strcasecmp (in_or_out, "in") == 0)
		interface->dlist_in[CONFIG_MRTD->protocol] = num;
	    else
		interface->dlist_out[CONFIG_MRTD->protocol] = num;
	}

	dlist = New (dlist_t);
	dlist->num = num;
	dlist->in_or_out = in_or_out;
	dlist->interface = NULL;
	LL_Add (rip->ll_dlists, dlist);
    }
    return (1);
}


static void
get_config_router_rip (rip_t *rip)
{
    int i;
    prefix_t *prefix;
    interface_t *interface;
    dlist_t *dlist;

    config_add_output ("router %s\n", proto2string (rip->proto));

    pthread_mutex_lock (&rip->mutex_lock);

    LL_Iterate (rip->ll_networks, prefix) {
	config_add_output ("  network %s\n", prefix_toax (prefix));
    }

    LL_Iterate (rip->ll_interfaces, interface) {
	config_add_output ("  network %s\n", interface->name);
    }

    LL_Iterate (rip->ll_dlists, dlist) {
        config_add_output ("  distribute-list %d %s %s\n",
                           dlist->num, dlist->in_or_out,
                           dlist->interface ? dlist->interface->name : "");
    }

    pthread_mutex_unlock (&rip->mutex_lock);

    pthread_mutex_lock (&MRT->mutex_lock);
    for (i = 0; i <= PROTO_MAX; i++) {
        if (BIT_TEST (MRT->redist[i], (1 << rip->proto)))
	    config_add_output ("  redistribute %s\n", proto2string (i));
    }
    pthread_mutex_unlock (&MRT->mutex_lock);
}


#ifdef HAVE_IPV6
/*
 * start RIPNG protocol configuration 
 */
static int
config_router_ripng (uii_connection_t * uii)
{
    if (running && !BIT_TEST (MRT->protocols, 1 << PROTO_RIPNG)) {
        config_notice (TR_ERROR, uii, "You can not do this while running\n");
	return (0);
    }
    CONFIG_MRTD->protocol = PROTO_RIPNG;
    uii->state = UII_CONFIG_ROUTER_RIPNG;
    config_notice (TR_TRACE, uii, "Router RIPng\n");
    if (BIT_TEST (MRT->protocols, 1 << PROTO_RIPNG))
	return (0);

    BIT_SET (MRT->protocols, 1 << PROTO_RIPNG);
    config_add_module (CF_DELIM, "router ripng", get_config_router_rip, RIPNG);
    return (1);
}
#endif /* HAVE_IPV6 */


/*
 * start RIP protocol configuration 
 */
static int
config_router_rip (uii_connection_t * uii)
{
    if (running && !BIT_TEST (MRT->protocols, 1 << PROTO_RIP)) {
        config_notice (TR_ERROR, uii, "You can not do this while running\n");
	return (0);
    }

    CONFIG_MRTD->protocol = PROTO_RIP;
    uii->state = UII_CONFIG_ROUTER_RIP;
    config_notice (TR_TRACE, uii, "Router RIP\n");
    if (BIT_TEST (MRT->protocols, 1 << PROTO_RIP))
	return (0);

    BIT_SET (MRT->protocols, 1 << PROTO_RIP);
    config_add_module (CF_DELIM, "router rip", get_config_router_rip, RIP);
    return (1);
}


void
config_rip_init (void)
{
    set_uii (UII, UII_PROMPT, UII_CONFIG_ROUTER_RIP, "Router RIP> ", 0);
#ifdef HAVE_IPV6
    set_uii (UII, UII_PROMPT, UII_CONFIG_ROUTER_RIPNG, "Router RIPng> ", 0);
#endif /* HAVE_IPV6 */

    /* Interactive commands */
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show rip", rip2_show,
		      "Show RIP status");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show rip routes",
		      rip2_show_routing_table, "Show RIP routing table");

#ifdef HAVE_IPV6
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show ripng", ripng_show,
		      "Show RIPng Status");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show ripng routes",
		      ripng_show_routing_table,
		      "Show RIPng routing table");
#endif /* HAVE_IPV6 */

    /* Config mode */
    uii_add_command2 (UII_CONFIG, COMMAND_NORM, "router rip",
		      config_router_rip, "Enables RIP routing protocol");
#ifdef HAVE_IPV6
    uii_add_command2 (UII_CONFIG, COMMAND_NORM, "router ripng",
		      config_router_ripng, "Enables RIPng routing protocol");
#endif /* HAVE_IPV6 */

    uii_add_command2 (UII_CONFIG_ROUTER_RIP, COMMAND_NORM, "network %p",
		      config_router_network_prefix, 
		      "Turns ON RIP routing for the network");
    uii_add_command2 (UII_CONFIG_ROUTER_RIP, COMMAND_NORM, "network %n",
		      config_router_network_interface, 
		      "Turns ON RIP routing on the interface");
#ifdef HAVE_IPV6
    uii_add_command2 (UII_CONFIG_ROUTER_RIPNG, COMMAND_NORM, "network %P",
		      config_router_network_prefix, 
		      "Turns ON RIPng routing for the network");
    uii_add_command2 (UII_CONFIG_ROUTER_RIPNG, COMMAND_NORM, "network %n",
		      config_router_network_interface, 
		      "Turns ON RIPng routing on the interface");
#endif /* HAVE_IPV6 */

    uii_add_command2 (UII_CONFIG_ROUTER_RIP, COMMAND_NORM,
	"redistribute (static|ospf|bgp|direct|connected|kernel)",
		      config_redistribute,
		      "Redistributes route from the protocol");
    uii_add_command2 (UII_CONFIG_ROUTER_RIP, COMMAND_NORM,
	"no redistribute (static|ospf|bgp|direct|connected|kernel)",
		      no_config_redistribute,
		      "Not redistribute route from the protocol");
#ifdef HAVE_IPV6
    uii_add_command2 (UII_CONFIG_ROUTER_RIPNG, COMMAND_NORM,
	"redistribute (static|ospf|bgp|direct|connected|kernel)",
		      config_redistribute,
		      "Redistributes route from the protocol");
    uii_add_command2 (UII_CONFIG_ROUTER_RIPNG, COMMAND_NORM,
	"no redistribute (static|ospf|bgp|direct|connected|kernel)",
		      no_config_redistribute,
		      "Not redistribute route from the protocol");
#endif /* HAVE_IPV6 */

    uii_add_command2 (UII_CONFIG_ROUTER_RIP, COMMAND_NORM,
		      "distribute-list %d (in|out) [%s]",
		      config_distribute_list,
		      "Applies access list to route [on interface]");
#ifdef HAVE_IPV6
    uii_add_command2 (UII_CONFIG_ROUTER_RIPNG, COMMAND_NORM,
		      "distribute-list %d (in|out) [%s]",
		      config_distribute_list,
		      "Applies access list to route [on interface]");
#endif /* HAVE_IPV6 */
}
