/* 
 * $Id: route_btoa.c,v 1.21 1998/09/04 20:44:48 masaki Exp $
 */

#include <mrt.h>

/* globals */
static int MACHINE_PARSEABLE_FLAG = 0;
static int MACHINE_PARSEABLE_FLAG2 = 0;
static char time_str[MAXLINE];

static char *sbgp_states[] =
{
    "NULL",
    "Idle",
    "Connect",
    "Active",
    "Opensent",
    "Openconfirm",
    "Established"
};

static io_t *IO;
static trace_t *default_trace;

void process_input ();

static void 
bgp_process_sync_msg (enum MRT_MSG_TYPES type, u_char *msg, int length, 
		      int *viewno, char *filename)
{
    *viewno = 0;
    BGP_GET_SHORT (*viewno, msg);
    length -= 2;
    strncpy (filename, msg, length);
    filename[length] = '\0';
}


static void 
bgp_process_state_msg (enum MRT_MSG_TYPES type, u_char *msg, int length, 
		       gateway_t ** gateway,
		       u_short * old_state, u_short * new_state)
{
    u_char *cp, ip[16];
    u_short as;
    prefix_t *prefix;
    int plen = 4;
    int family = AF_INET;
    gateway_t *gateway_from, *gateway_to;

#ifdef HAVE_IPV6
    if (type == MSG_PROTOCOL_BGP4PLUS || type == MSG_PROTOCOL_BGP4PLUS_01) {
	plen = 16;
        family = AF_INET6;
    }
#endif /* HAVE_IPV6 */

    cp = msg;

    /* from */
    BGP_GET_SHORT (as, cp);
    memcpy (ip, cp, plen);
    cp += plen;
    prefix = New_Prefix (family, ip, plen * 8);
    gateway_from = add_gateway (prefix, as, NULL);
    Deref_Prefix (prefix);

#ifdef notdef
    /* to */
    BGP_GET_SHORT (as, cp);
    memcpy (ip, cp, plen);
    cp += plen;
    prefix = New_Prefix (family, ip, plen * 8);
    gateway_to = add_gateway (prefix, as, NULL);
    Deref_Prefix (prefix);
#endif

    BGP_GET_SHORT (*old_state, cp);
    BGP_GET_SHORT (*new_state, cp);

    *gateway = gateway_from;
    return;
}


static void 
print_bgp_msg (mrt_msg_t * msg)
{
    gateway_t *gateway_to;
    bgp_attr_t *p_attr;
    LINKED_LIST *ll_with_prefixes, *ll_ann_prefixes;
    u_short *pref;
    u_short old_state, new_state;

    p_attr = NULL;
    ll_with_prefixes = NULL;
    ll_ann_prefixes = NULL;
    pref = NULL;

    if (msg->subtype == MSG_BGP_UPDATE) {
	bgp_process_update_msg (msg->type, msg->value, msg->length,
				&gateway_to,
				&p_attr,
				&ll_with_prefixes,
				&ll_ann_prefixes);
    }
    else if (msg->subtype == MSG_BGP_STATE_CHANGE) {
	if (!MACHINE_PARSEABLE_FLAG) {
	    bgp_process_state_msg (msg->type, msg->value, msg->length,
				   &gateway_to, &old_state, &new_state);
	    printf ("PEER: AS%d %s\n", gateway_to->AS, 
				       prefix_toa (gateway_to->prefix));
	    printf ("STATE: %s/%s\n", sbgp_states[old_state], 
				      sbgp_states[new_state]);
	}
	else {
	  bgp_process_state_msg (msg->type, msg->value, msg->length,
				 &gateway_to, &old_state, &new_state);
	  printf ("%s|STATE|%s|%d|%d|%d\n",
		  time_str, prefix_toa (gateway_to->prefix), 
		  gateway_to->AS, old_state, new_state);
	}
	return;
    }
    else if (msg->subtype == MSG_BGP_SYNC) {
	if (!MACHINE_PARSEABLE_FLAG) {
	    int viewno;
	    char filename[MAXLINE];
	    bgp_process_sync_msg (msg->type, msg->value, msg->length,
				  &viewno, filename);
	    printf ("VIEW: %d\n", viewno);
	    printf ("FILE: %s\n", filename);
	}
	return;
    }
    else {
	fprintf (stderr, "*** ERROR *** Unknown message subtype: %d\n", 
		 msg->subtype);
	return;
    }

    if (MACHINE_PARSEABLE_FLAG) {
	prefix_t *prefix;
	char *attr_string;

	if (ll_ann_prefixes) {
	    attr_string = bgp_attr_toa (p_attr);
	    LL_Iterate (ll_ann_prefixes, prefix) {
		if (MACHINE_PARSEABLE_FLAG2) {
		    char *cp = attr_string;
		    if ((cp = strchr (cp, '|')) != NULL &&
			(cp = strchr (cp+1, '|')) != NULL)
			*cp = '\0';
		    printf ("%s|A|%s|%d|%s|%s\n",
			    time_str, prefix_toa (p_attr->gateway->prefix), 
			    p_attr->gateway->AS,
			    prefix_toax (prefix), attr_string);
		}
		else {
		    printf ("%s|A|%s|%d|%s|%s\n",
			    time_str, prefix_toa (p_attr->gateway->prefix), 
			    p_attr->gateway->AS,
			    prefix_toax (prefix), attr_string);
		}
	    }
	    LL_Destroy (ll_ann_prefixes);
	}
	if (ll_with_prefixes) {
	    LL_Iterate (ll_with_prefixes, prefix) {
		printf ("%s|W|%s|%d|%s\n",
		        time_str, prefix_toa (p_attr->gateway->prefix), 
		        p_attr->gateway->AS, prefix_toax (prefix));
	    }
	    LL_Destroy (ll_with_prefixes);
	}

	if (p_attr)
	  bgp_deref_attr (p_attr);
	return;
    }

    if (ll_ann_prefixes == NULL && ll_with_prefixes == NULL)
	return;

    printf ("TO: AS%d   %s\n", gateway_to->AS, prefix_toa (gateway_to->prefix));
    /* print attributes in both cases
       because sometimes withdraw has some, which is strange, though */
    if (p_attr) {
      bgp_print_attr (p_attr);
      bgp_deref_attr (p_attr);
    }

    if (ll_ann_prefixes) {
	printf ("ANNOUNCE\n");
	if (msg->subtype == MSG_BGP_PREF_UPDATE)
	    print_pref_prefix_list (ll_ann_prefixes, pref);
	else
	    print_prefix_list (ll_ann_prefixes);
	LL_Destroy (ll_ann_prefixes);
    }
    if (ll_with_prefixes) {
	printf ("WITHDRAW\n");
	print_prefix_list (ll_with_prefixes);
	LL_Destroy (ll_with_prefixes);
    }
}





static void
print_msg_hdr (mrt_msg_t *msg)
{
    char *stime, **cpp;

    if (!MACHINE_PARSEABLE_FLAG) {
	static int first = 1;
	if (first) {
	    first = 0;
	}
	else {
	    printf("\n");
	}
	stime = my_strftime (msg->tstamp, "%D %T");
	printf ("TIME: %s\n", stime);
	printf ("TYPE: %s", S_MRT_MSG_TYPES[msg->type]);
	if ((cpp = S_MRT_MSG_SUBTYPES[msg->type]) != NULL)
	    printf ("/%s", cpp[msg->subtype]);
	printf("\n");
	Delete (stime);
    }
    else {
	if (MACHINE_PARSEABLE_FLAG2) {
	    char *cp;
            strcpy (time_str, S_MRT_MSG_TYPES[msg->type]);
	    /* BGP4+(1) -> BGP4+ */
	    if ((cp = strchr (time_str, '(')) != NULL)
		*cp = '\0';
	    strcat (time_str, "|");
	    cp = my_strftime (msg->tstamp, "%D %T");
	    strcat (time_str, cp);
	    Delete (cp);
	}
	else {
	    /* leave a message in time_str to use later */
            sprintf (time_str, "%s|%lu", S_MRT_MSG_TYPES[msg->type],
		         (u_long) msg->tstamp);
	}
    }
}

void 
main (int argc, char *argv[])
{
  char c;
    extern char *optarg;	/* getopt stuff */ 
    extern int optind;		/* getopt stuff */ 
    char *usage = "Usage: route_btoa [-m] [(-i|-r)] [input_binary_file(s)] \n"; 
    int errs; 
    int input_mode = 0;
    
    default_trace = New_Trace ();

    init_mrt (default_trace);
    IO = New_IO (default_trace);
    io_set (IO, IO_INFILE, "stdin", NULL);

    while ((c = getopt (argc, argv, "if:o:mMvh")) != -1)
      switch (c) {
	/* machine parseable format */
      case 'M':
	MACHINE_PARSEABLE_FLAG2 = 1;
      case 'm':
	MACHINE_PARSEABLE_FLAG = 1;
	break;
      case 'i':
      case 'f':
	input_mode = 1;
	break;
      case 'r':
	if (io_set (IO, IO_INMSGQ, (char *) (optarg), NULL) < 0) {
	  if (optarg) 
	    fprintf (stderr, "\nError setting infile %s\n", optarg);
	  else
	    fprintf (stderr, "\nError -- no infile given!\n");
	  errs++;
	}
	input_mode = 2;
	break;
      case 'v':
	set_trace (default_trace, TRACE_FLAGS, TR_ALL,
		   TRACE_LOGFILE, "stdout", NULL);
	break;
      case 'h':
	fprintf (stderr, usage);
	fprintf (stderr, "Machine output (-m flag):\n");
	fprintf (stderr, "Protocol|Time|Type|PeerIP|PeerAS|Prefix|<update dependant>\n");
	fprintf (stderr, "  (for BGP announcements) ASPATH|Origin|NextHop|LocPref|MED|");
	fprintf (stderr, "Community|AtomicAGG|Aggregator\n");
	exit (0);
      default:
	fprintf (stderr, usage);
	printf ("\nMRT version (%s) compiled on %s\n\n",
		MRT_VERSION, __DATE__);
	exit (0);
      }

    
    if (input_mode == 1) {
      for ( ; optind < argc; optind++) {
	if (io_set (IO, IO_INFILE, (char *) argv[optind], NULL) < 0) {
	  if (optarg) 
	    printf ("Failed to open infile %s", (char *) optarg);
	  else
	    printf ("Failed to open -- no file provided\n");
	  exit (1);
	}
	process_input ();
      }
    }
    else if (input_mode == 2) {
	process_input ();
    }
    else if (optind < argc) {
      for ( ; optind < argc; optind++) {
	if (io_set (IO, IO_INFILE, (char *) argv[optind], NULL) < 0) {
	  printf ("Failed to open infile %s", (char *) argv[optind]);
	  exit (1);
	}
	process_input ();
      }
    }
    else {
	if (io_set (IO, IO_INFILE, "stdin", NULL) >= 0)
	    process_input ();
    }
    exit (0);
}

void process_input () {
  mrt_msg_t *msg;
  
  while (1) {
    if ((msg = (mrt_msg_t *) io_read (IO)) == NULL) {
      /*       printf("\nInvalid MSG "); */
      return;
    }

    switch (msg->type) {
    case MSG_NULL:
      print_msg_hdr (msg);
      break;

    case MSG_START:
    case MSG_I_AM_DEAD:
    case MSG_PEER_DOWN:
    case MSG_PROTOCOL_IDRP:
    case MSG_PROTOCOL_RIPNG:
    case MSG_PROTOCOL_RIP:
      print_msg_hdr (msg);
      printf ("*** Not yet implemented ***\n");
      break;

    case MSG_DIE:
      print_msg_hdr (msg);
      exit(0);
      
    case MSG_PROTOCOL_BGP:
#ifdef HAVE_IPV6
    case MSG_PROTOCOL_BGP4PLUS:
    case MSG_PROTOCOL_BGP4PLUS_01:
#endif /* HAVE_IPV6 */
      print_msg_hdr (msg);
      print_bgp_msg (msg);
      break;
    case MSG_TABLE_DUMP:
      print_routing_table_msg (msg);
      break;
    default:
      printf ("*** Unknown message type %d\n", msg->type);
      break;
    }
    Delete (msg);
  }


  return;
}
